using System;
using System.IO;
using GTA;
using GTA.Native;

public class IV_CE_OvertakeFix : Script
{
	private const float TRUCK_DOT_PRODUCT_THRESHOLD = -0.96f;
    private const float CAR_DOT_PRODUCT_THRESHOLD = -0.97f;
    private const float BIKE_DOT_PRODUCT_THRESHOLD = -0.94f; 
	
    private const float CAR_DETECTION_DISTANCE = 6.8f;
    private const float TRUCK_DETECTION_DISTANCE = 9.2f;
    private const float BIKE_DETECTION_DISTANCE = 3.7f;

    private const int STOP_DURATION = 15000; 

    private Vehicle lastStoppedTruck;
    private DateTime lastTruckStoppedTime;
    private bool playerMovedSinceTruckUnfreeze = true;

    private Vehicle lastStoppedCar;
    private DateTime lastCarStoppedTime;
    private bool playerMovedSinceCarUnfreeze = true;

    private Vehicle lastStoppedBike;
    private DateTime lastBikeStoppedTime;
    private bool playerMovedSinceBikeUnfreeze = true;
	
	private bool canFreezeNextVehicle = true; 

    private readonly uint[] truckHashes = 
    {
		4011757448u,	// avan
		850991848u,		// biff
		3581397346u,	// bus
		2287941233u,	// pbus
		1938952078u,	// firetruk
		1353720154u,	// flatbed
		904750859u,		// mule
		569305213u,		// packer
		207497487u,		// packer2
		2157618379u,	// phantom
		1747439474u,	// stockade
		1911513875u,	// nstockade
		2394394458u,	// pstockade
		1917016601u,	// trash
		3195007082u,	// yankee
		2396846736u		// yankee2
    };

    private readonly uint[] bikeHashes = 
    {
		1672195559u,	// akuma
		3723957976u,	// angel
		4180675781u,	// bati
		3403504941u,	// bati2
		2464508460u,	// bobber
		2006142190u,	// daemon
		3886915065u,	// diabolus
		2623969160u,	// double
		2535109211u,	// double2
		2452219115u,	// faggio
		55628203u,		// faggio2
		1265391242u,	// hakuchou
		4039289119u,	// hakuchou2
		584879743u,		// hellfury
		301427732u,		// hexer
		4135840458u,	// innovation
		802082487u,		// lycan
		2688780135u,	// nightblade
		1203311498u,	// nrg900
		3385765638u,	// pcj
		4260343491u,	// policeb
		3935799761u,	// revenant
		788045382u,		// sanchez
		4154065143u,	// vader
		4217198264u,	// wayfarer
		3676349299u,	// wolfsbane
		3724934023u		// zombieb
	};

    public IV_CE_OvertakeFix()
    {
        Interval = 100;
        Tick += IV_CE_OvertakeFix_Tick;
    }
    
    private void IV_CE_OvertakeFix_Tick(object sender, EventArgs e)
    {
        try
        {
            if (!Player.Character.isInVehicle()) return;

            Vehicle playerVeh = Player.Character.CurrentVehicle;
            if (playerVeh == null || !playerVeh.Exists()) return;

            bool playerStopped = playerVeh.Speed < 0.1f;

            if (!playerStopped)
            {
                playerMovedSinceTruckUnfreeze = true;
                playerMovedSinceCarUnfreeze = true;
                playerMovedSinceBikeUnfreeze = true;

                lastTruckStoppedTime = DateTime.Now;
                lastCarStoppedTime = DateTime.Now;
                lastBikeStoppedTime = DateTime.Now;

                if (!canFreezeNextVehicle)
                {
                    canFreezeNextVehicle = true;
                }
            }

            int recognitionDelay = playerStopped ? 40000 : 1;

            if (playerStopped)
			{
				Vehicle vehBehind = FindVehicleBehind(playerVeh, recognitionDelay);
				if (vehBehind != null)
				{
					LogAndStopVehicle(vehBehind, playerStopped);
				}
			}
			UnfreezeVehiclesIfNeeded(playerStopped);
        }
		catch (Exception ex)
		{ }
    }

    private Vehicle FindVehicleBehind(Vehicle playerVeh, int recognitionDelay)
    {
        Vehicle vehBehind = null;
        float minDistance = float.MaxValue;
        Vector3 playerPosition = playerVeh.Position;
        Vector3 playerDirection = playerVeh.Direction;

        foreach (Vehicle veh in World.GetAllVehicles())
        {
            if (IsValidVehicle(veh, playerVeh, recognitionDelay))
            {
                float distance = (veh.Position - playerPosition).Length();
                if (distance < minDistance && IsInDetectionRange(veh, playerVeh, distance, playerPosition, playerDirection))
                {
                    vehBehind = veh;
                    minDistance = distance;
                }
            }
        }
		return vehBehind;
    }

    private bool IsValidVehicle(Vehicle veh, Vehicle playerVeh, int recognitionDelay)
    {
        if (veh == playerVeh || !veh.Exists()) return false;

        if (IsTruck(veh.Model.Hash))
        {
            if (veh == lastStoppedTruck && (DateTime.Now - lastTruckStoppedTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceTruckUnfreeze) return false;
        }
        else if (IsBike(veh.Model.Hash))
        {
            if (veh == lastStoppedBike && (DateTime.Now - lastBikeStoppedTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceBikeUnfreeze) return false;
        }
        else
        {
            if (veh == lastStoppedCar && (DateTime.Now - lastCarStoppedTime).TotalMilliseconds < recognitionDelay)
                return false;
            if (!playerMovedSinceCarUnfreeze) return false;
        }
        return true;
    }

    private bool IsInDetectionRange(Vehicle veh, Vehicle playerVeh, float distance, Vector3 playerPosition, Vector3 playerDirection)
    {
        float maxDetectionDistance;
        if (IsTruck(veh.Model.Hash))
        {
            maxDetectionDistance = TRUCK_DETECTION_DISTANCE;
        }
        else if (IsBike(veh.Model.Hash))
        {
            maxDetectionDistance = BIKE_DETECTION_DISTANCE;
        }
        else
        {
            maxDetectionDistance = CAR_DETECTION_DISTANCE;
        }
        return distance <= maxDetectionDistance && IsInDirection(playerVeh, veh, playerPosition, playerDirection);
    }

    private bool IsInDirection(Vehicle playerVeh, Vehicle veh, Vector3 playerPosition, Vector3 playerDirection)
    {
        Vector3 toVehicle = veh.Position - playerPosition;
        float toVehicleMagnitudeSquared = toVehicle.X * toVehicle.X + toVehicle.Y * toVehicle.Y;
        if (toVehicleMagnitudeSquared < 0.0001f) return false;

        float dotProduct = (playerVeh.Direction.X * toVehicle.X + playerVeh.Direction.Y * toVehicle.Y) / (float)Math.Sqrt(toVehicleMagnitudeSquared);
        float threshold;
        if (IsTruck(veh.Model.Hash))
        {
            threshold = TRUCK_DOT_PRODUCT_THRESHOLD;
        }
        else if (IsBike(veh.Model.Hash))
        {
            threshold = BIKE_DOT_PRODUCT_THRESHOLD;
        }
        else
        {
            threshold = CAR_DOT_PRODUCT_THRESHOLD;
        }
        return dotProduct <= threshold;
    }
	
	private bool IsTruck(int modelHash)
    {
        foreach (uint hash in truckHashes)
        {
            if (hash == modelHash)
                return true;
        }
        return false;
    }

    private bool IsBike(int modelHash)
    {
		foreach (uint hash in bikeHashes)
        {
            if (hash == modelHash)
			{
                return true;
			}
        }
        return false;
    }

    private void LogAndStopVehicle(Vehicle vehBehind, bool playerStopped)
    {
        Ped driver = vehBehind.GetPedOnSeat(VehicleSeat.Driver);
        if (driver != null && driver.Exists() && driver != Player.Character && playerStopped)
        {
            driver.Task.ClearAll();

            if (canFreezeNextVehicle)
            {
                if (IsTruck(vehBehind.Model.Hash))
				{
                    vehBehind.FreezePosition = true;
                    lastStoppedTruck = vehBehind;
                    lastTruckStoppedTime = DateTime.Now;
                }
                else if (IsBike(vehBehind.Model.Hash))
                {
                    driver.Task.StandStill(STOP_DURATION);
                    if (lastStoppedBike == null || !lastStoppedBike.Exists() || lastStoppedBike.Model.Hash != vehBehind.Model.Hash)
                    {
                        lastStoppedBike = vehBehind;
                        lastBikeStoppedTime = DateTime.Now;
                    }
                }
                else
                {
                    driver.Task.StandStill(STOP_DURATION);
                    if (lastStoppedCar == null || !lastStoppedCar.Exists() || lastStoppedCar.Model.Hash != vehBehind.Model.Hash)
                    {
                        lastStoppedCar = vehBehind;
                        lastCarStoppedTime = DateTime.Now;
                    }
                }
            }
        }
    }

    private void UnfreezeVehiclesIfNeeded(bool playerStopped)
    {
        if (lastStoppedTruck != null && lastStoppedTruck.Exists())
        {
            double truckFrozenDuration = (DateTime.Now - lastTruckStoppedTime).TotalMilliseconds;
            if (!playerStopped || truckFrozenDuration >= STOP_DURATION)
            {
                int hash = lastStoppedTruck.Model.Hash;
                lastStoppedTruck.FreezePosition = false;
                lastStoppedTruck = null;
                playerMovedSinceTruckUnfreeze = false;
                canFreezeNextVehicle = false;
            }
        }

        if (lastStoppedCar != null && lastStoppedCar.Exists())
        {
            double carStoppedDuration = (DateTime.Now - lastCarStoppedTime).TotalMilliseconds;
            if (!playerStopped || carStoppedDuration >= STOP_DURATION)
            {
                int hash = lastStoppedCar.Model.Hash;
                Ped driver = lastStoppedCar.GetPedOnSeat(VehicleSeat.Driver);
                if (driver != null && driver.Exists())
                {
                    driver.Task.ClearAll();
                }
                lastStoppedCar = null;
                playerMovedSinceCarUnfreeze = false;
                canFreezeNextVehicle = false;
            }
        }

        if (lastStoppedBike != null && lastStoppedBike.Exists())
        {
            double bikeStoppedDuration = (DateTime.Now - lastBikeStoppedTime).TotalMilliseconds;
            if (!playerStopped || bikeStoppedDuration >= STOP_DURATION)
            {
                int hash = lastStoppedBike.Model.Hash;
                Ped driver = lastStoppedBike.GetPedOnSeat(VehicleSeat.Driver);
                if (driver != null && driver.Exists())
                {
                    driver.Task.ClearAll();
                }
                lastStoppedBike = null;
                playerMovedSinceBikeUnfreeze = false;
                canFreezeNextVehicle = false;
            }
        }
    }
}
